CON

' *****************************************************************************
' *                                                                           *
' *     AiChip Industries LCC Virtual Machine                   Kernel.Spin   *
' *                                                                           *
' *****************************************************************************

' <<KERNEL FOR LCC>>

CON

' <<INSERT HEADER>>
' <<FINISH HEADER>>

' <<IF DEBUG>>
' <<OR TEST>>
' <<OR SPIN>>
' <<OR OBJECT>>
' <<ELSE>>
' <<SET SPIN>>
' <<END IF>>

' <<LIST DEBUG    Target   : Debuggable Spin File>>
' <<LIST TEST     Target   : Testable Spin File>>
' <<LIST SPIN     Target   : Standalone Spin File>>
' <<LIST OBJECT   Target   : Embeddable Object>>

' <<LIST SOURCE   Source   : >>
' <<LIST DATE     Created  : >>

' <<LIST COMPILE  Compiler : >>
' <<LIST MAKESPIN MakeSpin : >>
' <<LIST KERNEL   Kernel   : >>

' <<LIST OPTIONS  Options  : >>

' <<IF DEBUG>>
' <<OR TEST>>
' <<OR SPIN>>

CON

' *****************************************************************************
' *                                                                           *
' *     Configuration                                                         *
' *                                                                           *
' *****************************************************************************

CON

' .---------------------------------------------------------------------------.
' |     Define Board Configuration                                            |
' `---------------------------------------------------------------------------'

' <<INSERT CLKMODE>>

  _CLKMODE      = XTAL1+PLL16x
  _XINFREQ      = 7_372_800                             ' 118MHz Operation !

' <<FINISH CLKMODE>>

' <<END IF>>

' <<IF DEBUG>>
' <<OR TEST>>

CON

' .---------------------------------------------------------------------------.
' |     Define Hardware Configuration                                         |
' `---------------------------------------------------------------------------'

' <<INSERT TVPIN>>

  TV_PIN        = 12

' <<FINISH TVPIN>>

' <<END IF>>

' <<IF DEBUG>>

OBJ

' .---------------------------------------------------------------------------.
' |     Uses AiChip_CogDebug for debugging the Cog                            |
' `---------------------------------------------------------------------------'

' <<INSERT COGDEBUG>

  debug         : "AiChip_CogDebug_006"

' <<FINISH COGDEUG>>

' <<END IF>>

' <<IF DEBUG>>
' <<OR TEST>>

OBJ

' .---------------------------------------------------------------------------.
' |     Uses TV Display For debug reports                                     |
' `---------------------------------------------------------------------------'

  tv            : "TV_Text"

' <<END IF>>

' <<IF DEBUG>>
' <<OR TEST>>

CON

' *****************************************************************************
' *                                                                           *
' *     Main Program                                                          *
' *                                                                           *
' *****************************************************************************

' <<END IF>>

' <<IF KERNEL.SPIN>>

CON

' .---------------------------------------------------------------------------.
' |     Timing Test                                                           |
' `---------------------------------------------------------------------------'

PUB Kernel_TimingTest | varPtr

  ' Pasm :  28s  3,571,428 increments per second        1:1
  ' Spin : 640s    156,250 increments per second       23:1
  ' Lcc  : 510s    196,078 increments per second       18:1   byte reads
  ' Lcc  : 410s    243,900 increments per second       15:1   optimised reads
  ' Lcc  : 390s    256,410 increments per second       14:1   with branches
  ' Lcc  : 370s    270,270 increments per second       13:1   with branches to get var

  kernel_debug'@@@@@@@@@@@

  tv.Start( TV_PIN )

  tv.Str( Kernel_Version )
  if Kernel_BytecodeVersion
    tv.Out( "(" )
    tv.Str( Kernel_BytecodeVersion )
    tv.Out( ")" )
  if Kernel_SourceFileName
    tv.Str( String(" - ") )
    tv.Str( Kernel_SourceFileName )
  tv.Out( $0D )

  long[ varPtr := Kernel_GetVarAddress(0) ] := 0

  tv.Str( String("Watching : ") )
  tv.Hex( varPtr,4 )
  tv.Out( ":" )
  tv.Hex( varPtr-@LccCode,4 )
  case Kernel_GetVarSize(0)
    1 : tv.Str( String(" BYTE ") )
    2 : tv.Str( String(" WORD ") )
    4 : tv.Str( String(" LONG ") )
  tv.Str( Kernel_GetVarName(0) )
  tv.Out( $0D )
  tv.Out( $0D )

  Kernel_Start

  Debug.RunForever

  repeat while long[varPtr] < 10_000_000
    tv.Out( $0A )
    tv.Out(   0 )
    tv.Dec( long[varPtr] )

  tv.Str( String(" - FINISHED" ) )

' <<END IF>>

' <<SELECT CASE>>

' <<CASE DEBUG>>

CON

' .---------------------------------------------------------------------------.
' |     Debugging                                                             |
' `---------------------------------------------------------------------------'

PUB Kernel_Debug | xSp, xPc, xOp, xLocPtr, spCount

  tv.Start( TV_PIN )

  tv.Str( Kernel_Version )
  if Kernel_BytecodeVersion
    tv.Out( "(" )
    tv.Str( Kernel_BytecodeVersion )
    tv.Out( ")" )
  if Kernel_SourceFileName
    tv.Str( String(" - ") )
    tv.Str( Kernel_SourceFileName )
  tv.Str( String( $0D,"Debugging ...",$0D) )

  Kernel_Start

  repeat
    Debug.RunTo( Kernel_Label(@Fetch)+1 )

    ' Display stack
    if false
      xSp := Debug.Peek( Kernel_Label(@sp) )
      xLocPtr :=  Debug.Peek( Kernel_Label(@locPtr) )
      spCount := ( (xLocPtr-xSp) ~> 2 )+1
      ' Show sp / locPtr addresses
      if true
        tv.Hex( xSp,8)
        tv.Out( " " )
        tv.Hex( xLocPtr, 8)
        tv.Out( " " )
        tv.Dec( spCount)
        tv.Out( $0D)
      else
        tv.Dec( spCount )
        tv.Out( " ")
      if spCount > 0
        repeat spCount <# 3
          tv.Hex( long[ xSp ],8 )
          tv.Out( " " )
          xSp += 4
          spCount--
        if spCount > 0
          tv.Out( "+" )
        tv.Out($0D)

    ' Disassemble LCC bytecode to be executed        
    xPc := Debug.Peek( Kernel_Label(@pc) )
    xSp := Debug.Peek( Kernel_Label(@sp) )
    xOp := Debug.Peek( Kernel_Label(@OVERLAY_opc) )
    tv.Hex( xSp, 4 )
    tv.Out( " " )
    tv.Hex( xPc-@LccCode, 4 )
    tv.Out( " " )
    tv.Hex( xOp, 2 )
    tv.Out( " " )
    tv.Str( Kernel_FixOpcName( xPc, xOp, Kernel_GetOpcName(xOp) ) )
    tv.Out( $0D )

    ' Delay between steps
    if true
      WaitCnt( CLKFREQ / 1000 * 500 + CNT )

PRI Kernel_Label( atCogAdr )
  if atCogAdr => @OVERLAY_END
    return word[ atCogAdr ]
  return Debug.Label( atCogAdr )
  
PRI Kernel_FixOpcName( pcVal, opcVal, namPtr ) | i, n, digits
  digits := 4
  result := namPtr
  case opcVal & $F0
    $80   : opcVal &= 3
            if ( opcVal & $0C ) == $08 ' #ARG+
              opcVal <<= 2
            n := 0
            digits := 2 << opcVal
            repeat
              n := ( n << 8 ) | byte[++pcVal]
            while opcVal--
    $90   : n := ( ( opcVal & $0F ) << 28 ) ~> 28       ' #MEM+/#MEM-
            n := pcVal - @LccCode + n + 1
    $A0   : n := ( opcVal & $0F ) << 2                  ' #ARG+
    $B0   : n := opcVal & $0F                           ' #LOC+
    other :
      if opcVal < $80
        n := |< ( opcVal & $1F )
        if opcVal & $20
          n--
        if opcVal & $40
          n ^= $FFFF_FFFF
        digits := 8
      else
        return
  i := namPtr + StrSize(namPtr) - 8
  repeat 8
    byte[ i++ ] := Kernel_HexDigit( n >> 28 )
    n <<= 4
  byte[ i ] := 0

' <<CASE TEST>>

CON

' .---------------------------------------------------------------------------.
' |     Program Test                                                          |
' `---------------------------------------------------------------------------'

PUB Kernel_Testing | varCount, varNum, varPtr, n

  tv.Start( TV_PIN )

  tv.Str( Kernel_Version )
  if Kernel_BytecodeVersion
    tv.Out( "(" )
    tv.Str( Kernel_BytecodeVersion )
    tv.Out( ")" )
  if Kernel_SourceFileName
    tv.Str( String(" - ") )
    tv.Str( Kernel_SourceFileName )
  tv.Out( $0D )
  tv.Out( $0D )

  varCount := Kernel_GetVarCount <# 9

  repeat varNum from 0 To varCount-1
    varPtr := Kernel_GetVarAddress(varNum)
    tv.Hex( varPtr,4 )
    tv.Out( ":" )
    tv.Hex( varPtr-@LccCode,4 )
    case Kernel_GetVarSize(varNum)
      1 : tv.Str( String("  BYTE  ") )
      2 : tv.Str( String("  WORD  ") )
      4 : tv.Str( String("  LONG  ") )
    tv.Str( Kernel_GetVarName(varNum) )
    tv.Out( $0A )
    tv.Out(  30 )
    tv.Str( String( "       ", $0D ) )

  Kernel_Start

  repeat
    repeat varNum from 0 To varCount-1
      tv.Out( $0B )
      tv.Out( varNum+2 )
      tv.Out( $0A )
      varPtr := Kernel_GetVarAddress(varNum)
      case Kernel_GetVarSize(varNum)
        1 : tv.Out( 37 )
            tv.Hex( Kernel_PeekByte(varPtr), 2 )
        2 : tv.Out( 35 )
            tv.Hex( Kernel_PeekWord(varPtr), 4 )
        4 : tv.Out( 31 )
            tv.Hex( Kernel_PeekLong(varPtr), 8 )

PRI Kernel_PeekByte( varPtr )
  return byte[varPtr]

PRI Kernel_PeekWord( varPtr )
  return byte[varPtr] | ( byte[varPtr+1] << 8 )

PRI Kernel_PeekLong( varPtr )
  return Kernel_PeekWord(varPtr) | ( Kernel_PeekWord(varPtr+2) << 16 )

' <<END SELECT>>

' <<IF DEBUG>>
' <<OR TEST>>

' <<IF KERNEL.SPIN>>

CON

' *****************************************************************************
' *                                                                           *
' *     Support Utilities                                                     *
' *                                                                           *
' *****************************************************************************

' <<END IF>>

CON

' .---------------------------------------------------------------------------.
' |     Opcode Information                                                    |
' `---------------------------------------------------------------------------'

PRI Kernel_GetOpcCount : opcCount | opcPtr

  repeat while byte[opcPtr]
    opcCount++
    opcPtr++
    repeat while byte[opcPtr++]

PRI Kernel_GetOpcName( opcVal ) : namPtr

  if opcVal< $C0
    case opcVal & $F0
      $80   : if ( opcVal& $C0 )== 0
                namPtr := String( "NUM.K?  #$????????" )
              else
                namPtr := String( "NUM.K?? #???+$????????" )
                case opcVal & $0C
                  $4 : bytemove( namPtr+6, String("M #MEM"), 6 )
                  $8 : bytemove( namPtr+6, String("A #ARG"), 6 )
                  $C : bytemove( namPtr+6, String("L #LOC"), 6 )
              byte[namPtr+5] := ( opcVal & $03 ) + "1"
      $90   : namPtr := String( "NUM.MZ? #MEM+$????????" )
              byte[namPtr+6] := Kernel_HexDigit( opcVal & $0F )
      $A0   : namPtr := String( "NUM.AZ? #ARG+$????????" )
              byte[namPtr+6] := Kernel_HexDigit( opcVal & $0F )
      $B0   : namPtr := String( "NUM.LZ? #LOC+$????????" )
              byte[namPtr+6] := Kernel_HexDigit( opcVal & $0F )
      other : if opcVal
                namPtr := String( "NUM.KP? #$????????" )
                byte[namPtr+6] := ( ( opcVal & $60 ) >> 4 ) + "0"
              else
                namPtr := String( "BRK" )
  else
    namPtr := @OpcNames
    repeat while ( byte[namPtr] <> 0 ) and ( opcVal <> byte[namPtr] )
      namPtr++
      repeat while byte[namPtr++]
    namPtr++

PRI Kernel_HexDigit( n )
  if ( result := ( n & $F )+ "0" ) > "9"
    result += constant( "A"-"9"-1 )

CON

' .---------------------------------------------------------------------------.
' |        Variable Information                                               |
' `---------------------------------------------------------------------------'

PRI Kernel_GetVarCount : varCount | varPtr

  varPtr := @VarNames
  repeat while byte[varPtr]
    varCount++
    varPtr += 3
    repeat while byte[varPtr++]

PRI Kernel_GetVarName( varNum ) : varPtr

  varPtr := @VarNames
  repeat while varNum--
    varPtr += 3
    repeat while byte[varPtr++]
  varPtr += 3

PRI Kernel_GetVarAddress( varNum ) : varPtr

  varPtr := Kernel_GetVarName(varNum)
  varPtr := ( ( byte[varPtr-2] << 8 ) | byte[varPtr-1] ) + @LccCode

PRI Kernel_GetVarSize( varNum ) : varSize

  varSize := byte[Kernel_GetVarName(varNum)-3]

' <<END IF>

' <<IF DEBUG>
' <<OR TEST>>
' <<OR OBJECT>>

CON

' *****************************************************************************
' *                                                                           *
' *     Version Information                                                   *
' *                                                                           *
' *****************************************************************************

PUB Kernel_Version

  return String( "AiChip LccVm 002" )

PUB Kernel_BytecodeVersion

  ' <<INSERT BYTECODEVERSION>>

  return String( "??" )

  ' <<FINISH BYTECODEVERSION>>

PUB Kernel_SourceFileName

  ' <<INSERT SOURCEFILE>>

  return String( "Kernel.c" )

  ' <<FINISH SOURCEFILE>>

' <<END IF>>

' <<IF OBJECT>>

CON

' *****************************************************************************
' *                                                                           *
' *     Exported Interfaces                                                   *
' *                                                                           *
' *****************************************************************************

' <<INSERT METHODS>>

PUB Kernel_Main

  Kernel_Start

' <<FINISH METHODS>>

' <<END IF>>

CON

' *****************************************************************************
' *                                                                           *
' *     Kernel Launcher                                                       *
' *                                                                           *
' *****************************************************************************

VAR

  long  paramPc
  long  paramSp
  long  paramArgPtr
  long  paramLocPtr

' <<IF SPIN>>

PUB Kernel_AutoRun

  paramPc     := @LccCode

  paramSp     := $8000   - 4
  paramArgPtr := paramSp - 4
  paramLocPtr := paramSp - 4

  CogNew( @LccVm, @paramPc )

' <<ELSE>>

PUB Kernel_Start

  if paramPc == 0
    paramPc := @LccCode

  ' <<IF DEBUG>>
  ' <<OR TEST>>
  Kernel_SetStack(0,$6000)
  ' <<ELSE>>
  Kernel_SetStack(0,$8000)
  ' <<END IF>>

  Kernel_Push(0)
  long[ paramSp ] := 0

  ' <<IF DEBUG>>
  Debug.CogNew_WithDebugger( @LccVm, @paramPc, @DEBUG_BLOCK )
    ' <<IF KERNEL.SPIN>>
    return
    ' <<END IF>>
  ' <<ELSE>>
  CogNew( @LccVm, @paramPc )
  ' <<END IF>>

PUB Kernel_SetStack( argStackBase, argStackSize )

  if paramSp == 0
    paramSp     := argStackBase + argStackSize - 4
    paramArgPtr := paramSp
    paramLocPtr := paramSp

PUB Kernel_Push( argParameter )

  ' <<IF DEBUG>>
  ' <<OR TEST>>
  Kernel_SetStack(0,$6000)
  ' <<ELSE>>
  Kernel_SetStack(0,$8000)
  ' <<END IF>>

  long[ paramSp ] := argParameter

  paramSp     -= 4
  paramLocPtr -= 4

' <<END IF>>

' <<IF DEBUG>>

CON

' *****************************************************************************
' *                                                                           *
' *     Debug Support                                                         *
' *                                                                           *
' *****************************************************************************

DAT

DEBUG_BLOCK   long      0[3]

' <<END IF>>

CON

' *****************************************************************************
' *                                                                           *
' *     LCC Virtual Machine                                                   *
' *                                                                           *
' *****************************************************************************

DAT

' *****************************************************************************
' *                                                                           *
' *     Initialisation                                                        *
' *                                                                           *
' *****************************************************************************

'       If no arguments provided ...    If arguments provided ...
'
'                       .--------.                      .--------.
'       argPtr -------> |  "pc"  |      argPtr -------> | ARG+0  |
'                       |--------|                      | ARG+4  |
'       sp -------.---> |  zero  |                      | ARG+8  |
'       locPtr ---'     `--------'                      |--------|
'                                                       |  "pc"  |
'                                                       |--------|
'                                       sp -------.---> |  zero  |
'                                       locPtr ---'     `--------'
'
'       The "pc" is always zero - This is so that when the EXT executes it will
'       pull a zero address from the stack frame and we know we can halt and
'       return control to the caller. The caller can monitor "pc" and it will
'       be set to non-zero when it finished.

LccVm           org     $000

Lmm_Ret         jmp     #$+1

pc              call    #GetArg
sp              call    #getArg
argPtr          call    #GetArg
locPtr          call    #GetArg

memPtr          mov     memPtr,pc

tmp             jmp     #Fetch

' *****************************************************************************
' *                                                                           *
' *     Packed Number Handling ( $00 .. $7F )                                 *
' *                                                                           *
' *         .---------  Bit Number Indicator    n := 1 << b                   *
' *         |                                                                 *
' *     %0cdb bbbb                                                            *
' *       ||                                                                  *
' *       |`---------- Decrement Bit            n := n-1                      *
' *       |                                                                   *
' *       `----------- Inversion Bit            n := n ^ $FFFF_FFFF           *
' *                                                                           *
' *     Note that the Decrement is applied before the Inversion when both     *
' *     bits are set. Note that to get $0000_0001 the opcode used is not      *
' *     $00 but $21. This is forced by the compilation process and is done    *
' *     to keep $00 free as a special purpose opcode should it be needed.     *
' *                                                                           *
' *     The packing algorithm is slighty different to the one used in the     *
' *     Parallax Spin Interpreter bytecode but is the same principle.         *
' *                                                                           *
' *****************************************************************************
' *                                                                           *
' *     0cdbbbbb        000xxxxx      001xxxxx      010xxxxx      011xxxxx    *
' *                                                                           *
' *     xxx00000        00000001      00000000      FFFFFFFE      FFFFFFFF    *
' *     xxx00001        00000002      00000001      FFFFFFFD      FFFFFFFE    *
' *     xxx00010        00000004      00000003      FFFFFFFB      FFFFFFFC    *
' *     xxx00011        00000008      00000007      FFFFFFF7      FFFFFFF8    *
' *     xxx00100        00000010      0000000F      FFFFFFEF      FFFFFFF0    *
' *     xxx00101        00000020      0000001F      FFFFFFDF      FFFFFFE0    *
' *     xxx00110        00000040      0000003F      FFFFFFBF      FFFFFFC0    *
' *     xxx00111        00000080      0000007F      FFFFFF7F      FFFFFF80    *
' *     xxx01000        00000100      000000FF      FFFFFEFF      FFFFFF00    *
' *     xxx01001        00000200      000001FF      FFFFFDFF      FFFFFE00    *
' *     xxx01010        00000400      000003FF      FFFFFBFF      FFFFFC00    *
' *     xxx01011        00000800      000007FF      FFFFF7FF      FFFFF800    *
' *     xxx01100        00001000      00000FFF      FFFFEFFF      FFFFF000    *
' *     xxx01101        00002000      00001FFF      FFFFDFFF      FFFFE000    *
' *     xxx01110        00004000      00003FFF      FFFFBFFF      FFFFC000    *
' *     xxx01111        00008000      00007FFF      FFFF7FFF      FFFF8000    *
' *     xxx10000        00010000      0000FFFF      FFFEFFFF      FFFF0000    *
' *     xxx10001        00020000      0001FFFF      FFFDFFFF      FFFE0000    *
' *     xxx10010        00040000      0003FFFF      FFFBFFFF      FFFC0000    *
' *     xxx10011        00080000      0007FFFF      FFF7FFFF      FFF80000    *
' *     xxx10100        00100000      000FFFFF      FFEFFFFF      FFF00000    *
' *     xxx10101        00200000      001FFFFF      FFDFFFFF      FFE00000    *
' *     xxx10110        00400000      003FFFFF      FFBFFFFF      FFC00000    *
' *     xxx10111        00800000      007FFFFF      FF7FFFFF      FF800000    *
' *     xxx11000        01000000      00FFFFFF      FEFFFFFF      FF000000    *
' *     xxx11001        02000000      01FFFFFF      FDFFFFFF      FE000000    *
' *     xxx11010        04000000      03FFFFFF      FBFFFFFF      FC000000    *
' *     xxx11011        08000000      07FFFFFF      F7FFFFFF      F8000000    *
' *     xxx11100        10000000      0FFFFFFF      EFFFFFFF      F0000000    *
' *     xxx11101        20000000      1FFFFFFF      DFFFFFFF      E0000000    *
' *     xxx11110        40000000      3FFFFFFF      BFFFFFFF      C0000000    *
' *     xxx11111        80000000      7FFFFFFF      7FFFFFFF      80000000    *
' *                                                                           *
' *****************************************************************************

NUM_KP          mov     lhs,#1
                shl     lhs,opc
                test    opc,#$20        WC
        IF_C    sub     lhs,#1
                test    opc,#$40        WC
        IF_C    xor     lhs,k_FFFF_FFFF

' .---------------------------------------------------------------------------.
' |                                                                           |
' |     Push a result onto the stack. We put it here so we have a fairly      |
' |     quick and efficient route for packed number handling which falls      |
' |     straight back into the Fetch loop when completed.                     |
' |                                                                           |
' |     Note that the stack grows down from the top of memory so we need      |
' |     to decrement the Stack Pointer to push. Also note that the Stack      |
' |     pointer always points to top of stack. This makes it easier and       |
' |     quicker to deal with unary opcodes. The Stack remains aligned on      |
' |     long boundaries all the time; only longs are ever pushed to the       |
' |     stack.                                                                |
' |                                                                           |
' `---------------------------------------------------------------------------'

Push            sub     sp,#4
                wrlong  lhs,sp

' *****************************************************************************
' *                                                                           *
' *     Main Fetch Loop                                                       *
' *                                                                           *
' *****************************************************************************

Fetch           rdbyte  opc,pc          WZ
                add     pc,#1

                test    opc,#$80        WZ      ' Packed Number ( $00..$7F ) ?
        IF_Z    jmp     #NUM_KP                 ' Yes

                cmps    opc,#$90        WZ, WC  ' Other Opcodes ( $90..$FF ) ?
        IF_AE   jmp     #Others                 ' Yes

' *****************************************************************************
' *                                                                           *
' *     Multi-Byte Number Handling ( $80 .. $8F )                             *
' *                                                                           *
' *             .-----  How many bytes follow   00 = 1 - Byte                 *
' *             |                               01 = 2 - Word                 *
' *             |                               10 = 3 - Triple               *
' *             |                               11 = 4 - Long                 *
' *     %1000 ttss                                                            *
' *           |                                                               *
' *           `------  Adjustment Offset        00 = None                     *
' *                                             01 = MEM+                     *
' *                                             10 = ARG+                     *
' *                                             11 = LOC+                     *
' *                                                                           *
' *****************************************************************************

NUM_KX          mov     lhs,#0
                mov     ptr,opc
                and     ptr,#3          WZ

                rdbyte  lhs,pc
                add     pc,#1

        IF_Z    jmp     #:Done

:Loop           rdbyte  tmp,pc
                add     pc,#1
                shl     lhs,#8
                or      lhs,tmp

                djnz    ptr,#:Loop

' .---------------------------------------------------------------------------.
' |                                                                           |
' `---------------------------------------------------------------------------'

:Done           test    opc,#4          WC      '  C
                test    opc,#8          WZ      ' Z

  IF_Z_AND_C    add     lhs,memPtr              ' 01 = %1000 01ss    #MEM+n

  IF_Z          jmp     #Push                   ' 00 = %1000 00ss    #n

AdjustArgLoc

  IF_NC         shl     lhs,#2                  '                    #ARG+n

                neg     lhs,lhs

  IF_NC         add     lhs,argPtr              ' 10 = %1000 10ss    #ARG+n
  IF_C          add     lhs,locPtr              ' 11 = %1000 11ss    #LOC+n

                jmp     #Push

' *****************************************************************************
' *                                                                           *
' *                                                                           *
' *                                                                           *
' *****************************************************************************

Others          cmps    opc,#$C0        WZ, WC  ' Normal bytecode ( $C0..$FF ) ?
        IF_AE   jmp     #Dispatch               ' Yes

' *****************************************************************************
' *                                                                           *
' *     %1001 nnnn      NUM.MZX #MEM+n     Fast Load Memory   ( $90 .. $9F )  *
' *                                                                           *
' *     %1010 nnnn      NUM.AZX #ARG+n     Fast Load Argument ( $A0 .. $AF )  *
' *                                                                           *
' *     %1011 nnnn      NUM.LZX #LOC+n     Fast Load Local    ( $B0 .. $BF )  *
' *                                                                           *
' *****************************************************************************

NUM_XZX         mov     lhs,opc
                and     lhs,#$F

                test    opc,#$10        WC      '  C
                test    opc,#$20        WZ      ' Z

        IF_NZ   jmp     #AdjustArgLoc           ' 10 - 1010 nnnn AZX
                                                ' 11 - 1011 nnnn LZX + AZX

                shl     lhs,#28                 ' 01 - 1001 nnnn MZX
                sar     lhs,#28
                add     lhs,pc
                jmp     #Push

' *****************************************************************************
' *                                                                           *
' *     SPR Addressing                                                        *
' *                                                                           *
' *****************************************************************************

SPR_I32         rdbyte  lhs,pc
                add     pc,#1
                or      lhs,k_8000_01F0
                jmp     #Push

GetFromSpr      movs    GetfromSpr_Src,lhs
k_8000_0000     long    $8000_0000              ' NOP
GetFromSpr_Src  mov     lhs,0-0
GetFRomSpr_Ret  ret

PutToSpr        movd    PutToSpr_Dst,ptr
k_8000_01F0     long    $8000_01F0              ' NOP
PutToSpr_Dst    mov     0-0,lhs
                jmp     #Fetch
                
' *****************************************************************************
' *                                                                           *
' *     Opcode Dispatch ( $C0 .. $FF )                                        *
' *                                                                           *
' *****************************************************************************

Dispatch        sub     opc,#$C0
                add     opc,#OpcTable
  IF_ALWAYS     jmp     opc

' <<INSERT OPCTABLE>>

OpcTable        jmp     #JPI                    ' C0
                jmp     #ENT                    ' C1
                jmp     #INC_I32                ' C2

' <<FINISH OPCTABLE>>

' *****************************************************************************
' *                                                                           *
' *     Stack Handling                                                        *
' *                                                                           *
' *****************************************************************************

' .---------------------------------------------------------------------------.
' |                                                                           |
' `---------------------------------------------------------------------------'

PopLhsAndRhsAndPtr      rdlong  ptr,sp
                        add     sp,#4

PopLhsAndRhs            rdlong  rhs,sp
                        add     sp,#4

PopLhs                  rdlong  lhs,sp
                        add     sp,#4

PopLhs_Ret
PopLhsAndRhs_Ret
PopLhsAndRhsAndPtr_Ret  ret

' .---------------------------------------------------------------------------.
' |                                                                           |
' `---------------------------------------------------------------------------'

PopLhsThenPtr           rdlong  lhs,sp
                        add     sp,#4

PopPtr                  rdlong  ptr,sp
                        add     sp,#4
PopPtr_Ret
PopLhsThenPtr_Ret       ret

' *****************************************************************************
' *                                                                           *
' *     Variable Handling                                                     *
' *                                                                           *
' *****************************************************************************

' .---------------------------------------------------------------------------.
' |                                                                           |
' `---------------------------------------------------------------------------'

PopPtrGetByte           call    #PopPtr

GetByte                 shl     ptr,#1  WC, NR
              IF_C      call    #GetFromSpr
              IF_C      jmp     #GetByte_Ret
              
                        rdbyte  lhs,ptr
GetByte_Ret
PopPtrGetByte_Ret       ret

' .---------------------------------------------------------------------------.
' |                                                                           |
' `---------------------------------------------------------------------------'

PopPtrGetWord           call    #PopPtr

GetWord                 shl     ptr,#1  WC, NR
              IF_C      call    #GetFromSpr
              IF_C      jmp     #GetWord_Ret
              
                        test    ptr,#1  WZ

              IF_Z      rdword  lhs,ptr
              IF_Z      jmp     #GetWord_Ret

                        rdbyte  lhs,ptr
                        add     ptr,#1
                        rdbyte  tmp,ptr
                        shl     tmp,#8
                        or      lhs,tmp
                        sub     ptr,#1
GetWord_Ret
PopPtrGetWord_Ret       ret

' .---------------------------------------------------------------------------.
' |                                                                           |
' `---------------------------------------------------------------------------'

PopPtrGetLong           call    #PopPtr

GetLong                 shl     ptr,#1  WC, NR
              IF_C      call    #GetFromSpr
              IF_C      jmp     #GetLong_Ret
              
                        test    ptr,#3  WZ

              IF_Z      rdlong  lhs,ptr
              IF_Z      jmp     #GetLong_Ret

                        call    #GetWord

                        test    ptr,#1  WZ

              IF_NZ     jmp     #GetLongSlow

                        rdword  tmp,ptr
                        shl     tmp,#16
                        or      lhs,tmp
                        jmp     #GetLong_Ret

GetLongSlow             add     ptr,#2
                        rdbyte  tmp,ptr
                        add     ptr,#1
                        shl     tmp,#16
                        or      lhs,tmp
                        rdbyte  tmp,ptr
                        shl     tmp,#24
                        or      lhs,tmp
                        sub     ptr,#3
GetLong_Ret
PopPtrGetLong_Ret       ret

' *****************************************************************************
' *                                                                           *
' *     Debugging                                                             *
' *                                                                           *
' *****************************************************************************

' .---------------------------------------------------------------------------.
' |     Line Number Debugging                                                 |
' `---------------------------------------------------------------------------'

LIN             call    #PopLhs
                jmp     #Fetch

' *****************************************************************************
' *                                                                           *
' *     Subroutine and Function Calls                                         *
' *                                                                           *
' *****************************************************************************

' .---------------------------------------------------------------------------.
' |     Count calling argument prior to entering Subroutine or Function       |
' `---------------------------------------------------------------------------'

ARG_I32         add     argCnt,#4
                jmp     #Fetch

' .---------------------------------------------------------------------------.
' |     Jump to Subroutine ( Void Function - No return Value )                |
' |---------------------------------------------------------------------------|
' |     Jump to Function   ( Result returned on Stack )                       |
' `---------------------------------------------------------------------------'

'       Before the call ...             After the call ...
'
'                       |--------|                      |--------|
'       argPtr -------> | ARG+0  |                      |        |
'                       | ARG+4  |                      |        |
'                       |--------|                      |--------|
'                       | XXXXX  |                      |        |
'                       | XXXXX  |                      |        |
'                       | XXXXX  |                      |        |
'                       | XXXXX  |                      |        |
'                       |--------|                      |--------|
'       locPtr -------> | LOC+0  |                      |        |
'                       | LOC+4  |                      |        |
'                       | LOC+8  |                      |        |
'                       |--------|                      |--------|
'                       | PAR+0  |      argPtr -------> | ARG+0  |
'       sp -----------> | PAR+4  |                      | ARG+4  |
'                       `--------'                      |--------|
'                                                       | argPtr |
'                                                       | locPtr |
'                                                       | pc     |
'                                                       |--------|
'                                       sp / locPtr --> |  zero  | For ENT
'                                                       `--------'

JFI_I32         or      argPtr,#1               ' Indicate we want a result

JSI             add     argCnt,#4*4             ' Adjust arg count for the
                                                ' following four pushes

                sub     sp,#4                   ' Push the current state
                wrlong  argPtr,sp
                sub     sp,#4
                wrlong  locPtr,sp
                sub     sp,#4
                wrlong  pc,sp

                mov     locPtr,sp               ' Set the new local base
                sub     locPtr,#4

                mov     argPtr,locPtr           ' Set the new argument base
                add     argPtr,argCnt

                mov     argCnt,#0               ' No arguments pushed

                mov     lhs,#0                  ' Push a zero local var count
                jmp     #Push

' .---------------------------------------------------------------------------.
' |     Enter Subroutine ( Allocate Local Variables )                         |
' `---------------------------------------------------------------------------'

'       On entry ( no push ) ...        After completion ...
'
'                       |--------|                      |--------|
'       argPtr -------> | ARG+0  |      argPtr -------> | ARG+0  |
'                       | ARG+4  |                      | ARG+4  |
'                       |--------|                      |--------|
'                       | XXXXXX |                      | XXXXXX |
'                       | XXXXXX |                      | XXXXXX |
'                       | XXXXXX |      sp ------. -N   | XXXXXX |
'                       |--------|               |      |--------|
'       sp / locPtr --> |  zero  |      locPtr --|----> | LOC+0  |
'                       `--------'               |      | LOC+2  |
'                                                |      | LOC+4  |
'                                                `----> | LOC+8  |
'                                                       `--------'

'       On entry ( with push ) ...      After completion ...
'
'                       |--------|                      |--------|
'       argPtr -------> | ARG+0  |      argPtr -------> | ARG+0  |
'                       | ARG+4  |                      | ARG+4  |
'                       | ARG+4  |                      | ARG+4  |
'                       |--------|                      |--------|
'                       | XXXXXX |                      | XXXXXX |
'                       | XXXXXX |                      | XXXXXX |
'                       | XXXXXX |      sp ------. -N   | XXXXXX |
'                       |--------|               |      |--------|
'       locPtr -------> |  zero  |      locPtr --|----> | LOC+0  |
'       sp -----------> | number |               |      | LOC+2  |
'                       `--------'               |      | LOC+4  |
'                                                `----> | LOC+8  |
'                                                       `--------'
'
'       We know locPtr is always 4 below the stack frame so we don't care if
'       there's been a number pushed or not, we simply pop whatever's on the
'       top ( the number or the zero ) and then reset sp := locPtr+4. Having
'       done that we can then go about adjusting the stack for the locals.
'
'       Note that the size of locals is in bytes and may not be a multiple of
'       four ( longs ). This is because locals do not have long alignment.
'       this is not a problem because our stack stays long aligned and any
'       access into the locals area is done as normal for other variables so
'       that handles non-aligned variables as it usually does.

ENT             call    #PopLhs                 ' Get the size of locals

                mov     sp,locPtr               ' Restore sp
                add     sp,#4

                sub     sp,lhs                  ' Create space for locals

:Loop           test    sp,#3           WZ      ' Make sure sp aligns to long
        IF_NZ   sub     sp,#1
        IF_NZ   jmp     #:Loop

                jmp     #Fetch

' .---------------------------------------------------------------------------.
' |     Exit from Subroutine ( Void Function - No return Value )              |
' |---------------------------------------------------------------------------|
' |     Return from Function ( Leave Result on Stack )                        |
' `---------------------------------------------------------------------------'

'       Before the return ...           After the return ...
'
'                       |--------|                      |--------|
'                       | ARG+0  |      argPtr -------> | ARG+0  |
'                       | ARG+4  |                      | ARG+4  |
'                       |--------|                      |--------|
'                       | XXXXX  |                      | XXXXXX |
'                       | XXXXX  |                      | XXXXXX |
'                       | XXXXX  |                      | XXXXXX |
'                       |--------|                      |--------|
'                       | LOC+0  |      locPtr -------> | LOC+0  |
'                       | LOC+4  |                      | LOC+4  |
'                       | LOC+8  |                      | LOC+8  |
'                       |--------|                      |--------|
'                       | ?????  |                      | ?????  |
'                       | ?????  |  .-- sp -------.---> | ?????  |
'                       |--------|  |             |     |--------|
'       argPtr -------> | ARG+0  | -'             `---> | result |
'                       | ARG+4  |                      `--------'
'                       | ARG+8  |
'                       |--------|
'                       | argPtr |
'                       | locPtr |
'                       | pc     |
'                       |--------|
'       locPtr -------> | LOC+0  |
'       sp -------.---> | LOC+4  |
'                 |     |--------|
'                 `---> | result |      Result present if return from function
'                       `--------'
'
'       Note that we pop pc first. This is so we can detect whether we have a
'       real stack frame or a faked one created when the code was started. If
'       it's fake then pc will be zero ...
'
'       If no arguments provided ...    If arguments provided ...
'
'                       .--------.                      .--------.
'                       |  zero  |                      | ARG+0  |
'                       |--------|                      | ARG+4  |
'       locPtr -------> | ?????? |                      | ARG+8  |
'                       `--------'                      |--------|
'                                                       |  zero  |
'                                                       |--------|
'                                       locPtr -------> | ?????? |
'                                                       `--------'
'
RET_I32         call    #PopLhs

EXT             mov     sp,locPtr
                add     sp,#4

                rdlong  pc,sp           WZ      ' Restore PC

        IF_Z    jmp     #Halt                   ' Is zero so end of program

                add     sp,#4                   ' Restore locPtr
                rdlong  locPtr,sp

                mov     tmp,argPtr

                add     sp,#4                   ' Restore argPtr
                rdlong  argPtr,sp

                test    argPtr,#1       WC      ' Keep return value ?
                andn    argPtr,#1

                mov     sp,tmp                  ' Restore SP ( remove args )
                add     sp,#4

        IF_C    jmp     #Push                   ' Yes
                jmp     #Fetch                  ' No

' .---------------------------------------------------------------------------.
' |                                                                           |
' `---------------------------------------------------------------------------'

Halt
                jmp     #Halt

' *****************************************************************************
' *                                                                           *
' *     Conditional Execution                                                 *
' *                                                                           *
' *****************************************************************************

' .---------------------------------------------------------------------------.
' |     Jump Unconditionally                                                  |
' `---------------------------------------------------------------------------'

JPI             Call    #PopPtr
                mov     pc,ptr
                jmp     #Fetch

' .---------------------------------------------------------------------------.
' |     Jump Conditionally on Equality or Non-Equality                        |
' `---------------------------------------------------------------------------'

JEQ_I32         Call    #PopLhsAndRhsAndPtr
                cmp     lhs,rhs         WZ
        IF_Z    mov     pc,ptr
                jmp     #Fetch

JNE_I32         Call    #PopLhsAndRhsAndPtr
                cmp     lhs,rhs         WZ
        IF_NZ   mov     pc,ptr
                jmp     #Fetch

' .---------------------------------------------------------------------------.
' |     Jump Conditionally on Relative Values                                 |
' `---------------------------------------------------------------------------'

JGT_I32         Call    #PopLhsAndRhsAndPtr
                cmp     lhs,rhs         WZ, WC
        IF_A    mov     pc,ptr
                jmp     #Fetch

JGE_I32         Call    #PopLhsAndRhsAndPtr
                cmp     lhs,rhs         WZ, WC
        IF_AE   mov     pc,ptr
                jmp     #Fetch

JLT_I32         Call    #PopLhsAndRhsAndPtr
                cmp     lhs,rhs         WZ, WC
        IF_B    mov     pc,ptr
                jmp     #Fetch

JLE_I32         Call    #PopLhsAndRhsAndPtr
                cmp     lhs,rhs         WZ, WC
        IF_BE   mov     pc,ptr
                jmp     #Fetch

JGT_S32         Call    #PopLhsAndRhsAndPtr
                cmps    lhs,rhs         WZ, WC
        IF_A    mov     pc,ptr
                jmp     #Fetch

JGE_S32         Call    #PopLhsAndRhsAndPtr
                cmps    lhs,rhs         WZ, WC
        IF_AE   mov     pc,ptr
                jmp     #Fetch

JLT_S32         Call    #PopLhsAndRhsAndPtr
                cmps    lhs,rhs         WZ, WC
        IF_B    mov     pc,ptr
                jmp     #Fetch

JLE_S32         Call    #PopLhsAndRhsAndPtr
                cmps    lhs,rhs         WZ, WC
        IF_BE   mov     pc,ptr
                jmp     #Fetch

' *****************************************************************************
' *                                                                           *
' *     Structure Copying                                                     *
' *                                                                           *
' *****************************************************************************

MOV_I08         call    #PopLhsAndRhsAndPtr

:Loop           rdbyte  tmp,rhs
                add     rhs,#1
                wrbyte  tmp,lhs
                add     lhs,#1
                djnz    ptr,#:Loop
                jmp     #Fetch

' <<IF NORAM>>

' *****************************************************************************
' *                                                                           *
' *     ROM / Eeprom Handling                                                 *
' *                                                                           *
' *****************************************************************************

'       If we are running code from ROM we need a way to determine if an address
'       of #MEM+ refers to a global variable or to code. As there are no spare
'       bits in the opcodes the way we do this is by prefixing a #MEM+ to ROM
'       with a ROM opcode. We do this by setting a bit in memPtr above what it
'       would normally be so when it's added it can be vectored off elsewhere.
'       Once vectored we clear the bit so we're back pointing to variables.
'
ROM             or      memPtr,k_8000_0000
                jmp     #Fetch

' <<END IF>>

' *****************************************************************************
' *                                                                           *
' *     Variable Manipulation                                                 *
' *                                                                           *
' *****************************************************************************

' .---------------------------------------------------------------------------.
' |     Get a Variable's contents whose Address is popped from the Stack      |
' `---------------------------------------------------------------------------'

GTI_I08         call    #PopPtrGetByte
                jmp     #Push

GTI_I16         call    #PopPtrGetWord
                jmp     #Push

GTI_I32         call    #PopPtrGetLong
                jmp     #Push

GTI_S08         call    #PopPtrGetByte
                shl     lhs,#24
                sar     lhs,#24
                jmp     #Push

GTI_S16         call    #PopPtrGetWord
                shl     lhs,#16
                sar     lhs,#16
                jmp     #Push

' .---------------------------------------------------------------------------.
' |     Set a byte variable's contents whose Address is popped from the Stack |
' `---------------------------------------------------------------------------'

DEC_I08         neg     rhs,#1
                jmp     #$+2
INC_I08         mov     rhs,#1
                call    #PopPtrGetByte
                call    #AddSkip1

PTI_I08         call    #PopLhsThenPtr

PutByte         shl     ptr,#1          WC, NR
      IF_C      jmp     #PutToSpr
              
                wrbyte  lhs,ptr
                jmp     #Fetch

' .---------------------------------------------------------------------------.
' |     Set a word variable's contents whose Address is popped from the Stack |
' `---------------------------------------------------------------------------'

DEC_I16         neg     rhs,#1
                jmp     #$+2
INC_I16         mov     rhs,#1
                call    #PopPtrGetWord
                call    #AddSkip1

PTI_I16         call    #PopLhsThenPtr

PutWord         shl     ptr,#1          WC, NR
      IF_C      jmp     #PutToSpr
              
                test    ptr,#1          WZ, NR

        IF_Z    wrword  lhs,ptr
        IF_Z    jmp     #Fetch

PutWordSlow     wrbyte  lhs,ptr
                add     ptr,#1
                shr     lhs,#8
                wrbyte  lhs,ptr
                jmp     #Fetch

' .---------------------------------------------------------------------------.
' |     Set a long variable's contents whose Address is popped from the Stack |
' `---------------------------------------------------------------------------'

DEC_I32         neg     rhs,#1
                jmp     #$+2
INC_I32         mov     rhs,#1
                call    #PopPtrGetLong
                call    #AddSkip1

PTI_I32         call    #PopLhsThenPtr

PutLong         shl     ptr,#1          WC, NR
      IF_C      jmp     #PutToSpr
              
                test    ptr,#3          WZ, NR

        IF_Z    wrlong  lhs,ptr
        IF_Z    jmp     #Fetch

                test    ptr,#1          WZ, NR

        IF_Z    wrword  lhs,ptr
        IF_Z    jmp     #PutWord

                wrbyte  lhs,ptr
                add     ptr,#1
                shr     lhs,#8
                wrbyte  lhs,ptr
                add     ptr,#1
                shr     lhs,#8
                jmp     #PutWordSlow

                ' .-----------------------------------------------------------.
                ' |                                                           |
                ' `-----------------------------------------------------------'

AddSkip1        add     AddSkip1_Ret,#1
                add     lhs,rhs
AddSkip1_Ret    ret

' *****************************************************************************
' *                                                                           *
' *     Binary Operators                                                      *
' *                                                                           *
' *****************************************************************************

' .---------------------------------------------------------------------------.
' |     Standard Maths                                                        |
' `---------------------------------------------------------------------------'

ADD_I32         call    #PopLhsAndRhs
                add     lhs,rhs
                jmp     #Push

SUB_I32         call    #PopLhsAndRhs
                sub     lhs,rhs
                jmp     #Push

' .---------------------------------------------------------------------------.
' |     Bit-Wise Operations                                                   |
' `---------------------------------------------------------------------------'

AND_I32         call    #PopLhsAndRhs
                and     lhs,rhs
                jmp     #Push

IOR_I32         call    #PopLhsAndRhs
                or      lhs,rhs
                jmp     #Push

XOR_I32         call    #PopLhsAndRhs
                xor     lhs,rhs
                jmp     #Push

' .---------------------------------------------------------------------------.
' |     Shift Operations                                                      |
' `---------------------------------------------------------------------------'

SL2_I32         call    #PopLhs
                mov     rhs,#2
                jmp     #$+2
SHL_I32         call    #PopLhsAndRhs
                shl     lhs,rhs
                jmp     #Push

SR2_I32         call    #PopLhs
                mov     rhs,#2
                jmp     #$+2
SHR_I32         call    #PopLhsAndRhs
                shr     lhs,rhs
                jmp     #Push

ROL_I32         call    #PopLhsAndRhs
                rol     lhs,rhs
                jmp     #Push

ROR_I32         call    #PopLhsAndRhs
                ror     lhs,rhs
                jmp     #Push

ASR_I32         call    #PopLhsAndRhs
                sar     lhs,rhs
                jmp     #Push

' .---------------------------------------------------------------------------.
' |                                                                           |
' `---------------------------------------------------------------------------'

HGH_I32         call    #PopLhsAndRhs
                cmp     rhs,lhs         WZ, WC
        IF_A    mov     lhs,rhs
                jmp     #Push

LOW_I32         call    #PopLhsAndRhs
                cmp     rhs,lhs         WZ, WC
        IF_B    mov     lhs,rhs
                jmp     #Push

HGH_S32         call    #PopLhsAndRhs
                cmps    rhs,lhs         WZ, WC
        IF_A    mov     lhs,rhs
                jmp     #Push

LOW_S32         call    #PopLhsAndRhs
                cmps    rhs,lhs         WZ, WC
        IF_B    mov     lhs,rhs
                jmp     #Push

' *****************************************************************************
' *                                                                           *
' *     Unary Type Conversions                                                *
' *                                                                           *
' *****************************************************************************

' .---------------------------------------------------------------------------.
' |     Shrink Unsigned Integer Size                                          |
' `---------------------------------------------------------------------------'

I2I_I08         rdlong  lhs,sp
                and     lhs,#$FF
                jmp     #UnaryWriteBack

I2I_I16         rdlong  lhs,sp
                and     lhs,k_0000_FFFF
                jmp     #UnaryWriteBack

' .---------------------------------------------------------------------------.
' |     Shrink Signed Integer Size                                            |
' `---------------------------------------------------------------------------'

S2S_S08         rdlong  lhs,sp
                shl     lhs,#24
                sar     lhs,#24
                wrlong  lhs,sp
                jmp     #Fetch

S2S_S16         rdlong  lhs,sp
                shl     lhs,#16
                sar     lhs,#16
                wrlong  lhs,sp
                jmp     #Fetch

' *****************************************************************************
' *                                                                           *
' *     Unary Operators                                                       *
' *                                                                           *
' *****************************************************************************

' .---------------------------------------------------------------------------.
' |     Duplicate Top of Stack                                                |
' `---------------------------------------------------------------------------'

DUP_I32         rdlong  lhs,sp
                jmp     #Push

' .---------------------------------------------------------------------------.
' |     Drop the Top of Stack                                                 |
' `---------------------------------------------------------------------------'

DRP_I32         add     sp,#4
                jmp     #Fetch

' .---------------------------------------------------------------------------.
' |     Numeric Negation                                                      |
' `---------------------------------------------------------------------------'

NEG_I32         rdlong  lhs,sp
                neg     lhs,lhs
                jmp     #UnaryWriteBack

' .---------------------------------------------------------------------------.
' |     Get Absolute ( Positive ) Value                                       |
' `---------------------------------------------------------------------------'

ABS_I32         rdlong  lhs,sp
                abs     lhs,lhs
                jmp     #UnaryWriteBack

' .---------------------------------------------------------------------------.
' |     Bit-Wise Inversion                                                    |
' `---------------------------------------------------------------------------'

NOT_I32         rdlong  lhs,sp
                xor     lhs,k_FFFF_FFFF

UnaryWriteBack  wrlong  lhs,sp
                jmp     #Fetch

' *****************************************************************************
' *                                                                           *
' *     LMM Interpreter                                                       *
' *                                                                           *
' *****************************************************************************

LMM             call    #PopLhs
                mov     lmmPc,lhs
                
Lmm_Fetch       rdlong  Lmm_Opcode,lmmPc
                add     lmmPc,#4
Lmm_Opcode      mov     0-0,0-0
                jmp     #Lmm_Fetch

' *****************************************************************************
' *                                                                           *
' *     Constants                                                             *
' *                                                                           *
' *****************************************************************************

' <<DEFINE CONSTANTS>>

k_0000_FFFF     long    $0000_FFFF 
k_FFFF_FFFF     long    $FFFF_FFFF

' <<FINISH CONSTANTS>>

DAT

' *****************************************************************************
' *                                                                           *
' *     Virtual Machine Register                                              *
' *                                                                           *
' *****************************************************************************
 
' .---------------------------------------------------------------------------.
' |     Core registers which the LCC Virtual Machine depends upon             |
' `---------------------------------------------------------------------------'

argCnt          long    0
' <<IF NORAM>>
romPtr          long    0-0
' <<END IF>>

' .---------------------------------------------------------------------------.
' |     Core registers which the LMM execution handler depends upon           |
' `---------------------------------------------------------------------------'

lmmPc           long    0-0

' .---------------------------------------------------------------------------.
' |     Registers which the running LMM will require access to                |
' `---------------------------------------------------------------------------'

lhs             long    0-0
rhs             long    0-0
ptr             long    0-0

' .---------------------------------------------------------------------------.
' |     Initialisation routine pre-loaded into overlay area                   |
' `---------------------------------------------------------------------------'

INITIALISE         

GetArg          mov     $1F7,GetArg_Ret
                sub     $1F7,#1
                movd    :PtrGet,$1F7
                mov     $1F7,PAR
:PtrInc         add     $1F7,#0
:PtrGet         rdlong  0-0,$1F7
                add     :PtrInc,#4
GetArg_Ret      ret

INITIALISE_END

' .---------------------------------------------------------------------------.
' |     Ensure everything fits into the Cog                                   |
' `---------------------------------------------------------------------------'

' <<IF DEBUG>
                fit     $1EF
' <<ELSE>
                fit     $1F0
' <<END IF>>

' .---------------------------------------------------------------------------.
' |     Temporary registers                                                   |
' `---------------------------------------------------------------------------'

                org     INITIALISE

OVERLAY

opc             res     1

OVERLAY_END

' .---------------------------------------------------------------------------.
' |     Esure everything still fits into the Cog                              |
' `---------------------------------------------------------------------------'

' <<IF DEBUG>
                fit     $1EF
' <<ELSE>
                fit     $1F0
' <<END IF>>

' <<IF DEBUG>>

' .---------------------------------------------------------------------------.
' |     Register address support for debugger when using overlay registers    |
' `---------------------------------------------------------------------------'

OVERLAY_opc     word    opc

' <<END IF>>

CON

' *****************************************************************************
' *                                                                           *
' *     End Of LCC Virtual Machine                                            *
' *                                                                           *
' *****************************************************************************

' <<INSERT LMMCODE>>

CON

' *****************************************************************************
' *                                                                           *
' *     LMM Executed Library Code                                             *
' *                                                                           *
' *****************************************************************************

' <<FINISH LMMCODE>>

' <<IF DEBUG>>
' <<OR TEST>>

CON

' *****************************************************************************
' *                                                                           *
' *     Printable names of opcodes                                            *
' *                                                                           *
' *****************************************************************************

DAT

' <<INSERT OPCNAMES>>

OpcNames        byte    $C0, "JPI", 0
                byte    $C1, "ENT.LOZ", 0
                byte    $C2, "INC.I32", 0
                byte    $00, "?",0

' <<FINISH OPCNAMES>>

CON

' *****************************************************************************
' *                                                                           *
' *     End of Printable names of opcodes                                     *
' *                                                                           *
' *****************************************************************************

CON

' *****************************************************************************
' *                                                                           *
' *     Printable names of variables                                          *
' *                                                                           *
' *****************************************************************************

DAT

' <<INSERT VARNAMES>>

VarNames        byte    4, $00, $04, "count", 0
                byte    0

' <<FINISH VARNAMES>>

CON

' *****************************************************************************
' *                                                                           *
' *     End of Printable names of variables                                   *
' *                                                                           *
' *****************************************************************************

' <<END IF>>

CON

' *****************************************************************************
' *                                                                           *
' *     LCC Executable Code                                                   *
' *                                                                           *
' *****************************************************************************

DAT

' <<IF DEBUG>>
' <<OR TEST>>

        long                                    ' Align to long

' <<END IF>>
        
' <<INSERT LCCCODE>>

        
LccCode byte    $85, $00, $08                   ' 0000          NUM.KM2 #MEM+main
        byte    $C0                             ' 0003          JPI
        byte    $00, $00, $00, $00              ' 0004  count:  LNG     0
        byte    $C1                             ' 0008  main:   ENT
        byte    $84, $04                        ' 0009  @1:     NUM.KM1 #MEM+count
        byte    $C2                             ' 000B          INC.I32
        byte    $9C                             ' 000C  @2:     NUM.KMZ #MEM+@1
        byte    $C0                             ' 000D          JPI
                                                ' 000E          END

' <<FINISH LCCCODE>>

CON

' *****************************************************************************
' *                                                                           *
' *     End Of LCC Executable Code                                            *
' *                                                                           *
' *****************************************************************************
